#
# (C) Semantix Information Technologies.
#
# Semantix Information Technologies is licensing the code of the
# Data Modelling Tools (DMT) in the following dual-license mode:
#
# Commercial Developer License:
#       The DMT Commercial Developer License is the suggested version
# to use for the development of proprietary and/or commercial software.
# This version is for developers/companies who do not want to comply
# with the terms of the GNU Lesser General Public License version 2.1.
#
# GNU LGPL v. 2.1:
#       This version of DMT is the one to use for the development of
# applications, when you are willing to comply with the terms of the
# GNU Lesser General Public License version 2.1.
#
# Note that in both cases, there are no charges (royalties) for the
# generated code.
#
import re

from commonPy.asnAST import AsnInt, AsnMetaMember, AsnEnumerated, AsnString, AsnSetOf, AsnSequenceOf, AsnChoice, AsnReal, AsnBool
from commonPy.utility import panic, inform, warn
import commonPy.asnParser

import commonPy.configMT
from createInternalTypes import ScanChildren

g_output_C_HeaderFile = ""
g_declared_C_Types = {}
g_declared_Ada_Types = {}
g_choiceNodeTypename = ""


def CleanName(name):
    return re.sub(r'[^a-zA-Z0-9_]', '_', name)


def GetIntegerRangeAsAda(node):
    assert isinstance(node, AsnInt) or isinstance(node, AsnMetaMember)
    retValue = ""
    if len(node._range) == 1:
        retValue += " range "+str(node._range[0])+" .. "+str(node._range[0])  # pragma: no cover
    if len(node._range) == 2:
        if node._range[0] != -2147483648 or node._range[1] != 2147483647:
            retValue += " range "+str(node._range[0])+" .. "+str(node._range[1])
    return retValue


def MapType_to_Ada(isRoot, parentNode, node, leafTypeDict, names):
    if isinstance(node, AsnEnumerated):
        pass
    try:
        realLeafType = leafTypeDict[node._leafType]
    except:
        realLeafType = node._leafType
    if realLeafType == "INTEGER":
        return "Int64bits" + GetIntegerRangeAsAda(node)
    elif realLeafType == "REAL":
        return "Long_Float"
    elif realLeafType == "BOOLEAN":
        return "Bool32bits"
    elif realLeafType == "OCTET STRING":
        if isRoot:
            return "array (0..%d) of Character" % (int(node._range[-1])-1)
        else:
            return node._pseudoname
    elif realLeafType == "ENUMERATED":
        if len(node._members)>127:
            warn("ENUMERATED with more than 127 options (%s)... Untested configuration." % node.Location())  # pragma: no cover
        if not isRoot:
            return node._pseudoname
        else:  # pragma: no cover
            panic("MapType_to_Ada must not be called with root of ENUMERATED")  # pragma: no cover
    elif realLeafType == "SEQUENCEOF" or realLeafType == "SETOF":
        if isRoot:
            result = "array (0..%d) of " % (int(node._range[-1])-1)
            if isinstance(node._containedType, str):
                result += CleanName(node._containedType)
            else:  # Never called, due to the fixup at the end of VerifyAndFixAST                  pragma: no cover
                result += MapType_to_Ada(False, node, node._containedType, leafTypeDict, names)  # pragma: no cover
        else:  # this code will never be executed, because of the fixup in VerifyAndFixAST         pragma: no cover
            result = "TaStE_" + CleanName(parentNode[0][:1].upper() + parentNode[0][1:] + "_type")         # pragma: no cover
        return result
    elif realLeafType == "SEQUENCE" or realLeafType == "SET":
        result = "record\n"
        for x in node._members:
            if isinstance(x[1], AsnMetaMember):
                if leafTypeDict[x[1]._containedType] == "INTEGER":
                    if not hasattr(x[1], "_range"):
                        if hasattr(x[1], "_Min") and hasattr(x[1], "_Max"):
                            x[1]._range = [x[1]._Min, x[1]._Max]
                        else:  # pragma: no cover  (because of handling inside xmlASTtoAsnAST, all AsnMetaMember get _Min)
                            x[1]._range = []  # pragma: no cover (and _Max)
                    result += "    %s : %s%s;\n" % (CleanName(x[0]), CleanName(x[1]._containedType), GetIntegerRangeAsAda(x[1]))
                else:
                    result += "    %s : %s;\n" % (CleanName(x[0]), CleanName(x[1]._containedType))
            else:
                result += "    %s : %s;\n" % (CleanName(x[0]), MapType_to_Ada(False, x, x[1], leafTypeDict, names))
        #    if isinstance(x[1], AsnSequenceOf) or \
        #       isinstance(x[1], AsnSetOf) or \
        #       isinstance(x[1], AsnString) or \
        #       (isinstance(x[1], AsnMetaMember) and \
        #           (isinstance(names[x[1]._containedType], AsnSequenceOf) or \
        #            isinstance(names[x[1]._containedType], AsnSetOf) or \
        #            isinstance(names[x[1]._containedType], AsnString))):
        #       result += "    " + CleanName(x[0]) + "_len : Integer;\n"
        result += "end record"
        return result
    elif realLeafType == "CHOICE":
        global g_choiceNodeTypename
        if g_choiceNodeTypename == "":
            panic("Unhandled nested CHOICEs in dumpable.py. Please report this grammar to Semantix.")  # pragma: no cover
        result = "record\n"
        result += "    -- 1 means first field, 2 means second, etc\n"
        result += "    case choice_%s is\n" % CleanName(g_choiceNodeTypename)
        for i, x in enumerate(node._members):
            result += "      when %d =>\n" % (i+1)
            if isinstance(x[1], AsnMetaMember):
                if leafTypeDict[x[1]._containedType] == "INTEGER":
                    if not hasattr(x[1], "_range"):
                        if hasattr(x[1], "_Min") and hasattr(x[1], "_Max"):
                            x[1]._range = [x[1]._Min, x[1]._Max]
                        else:  # pragma: no cover  (because of handling inside xmlASTtoAsnAST, all AsnMetaMember get _Min)
                            x[1]._range = []  # pragma: no cover (and _Max)
                    result += "        %s : %s%s;\n" % (CleanName(x[0]), CleanName(x[1]._containedType), GetIntegerRangeAsAda(x[1]))
                else:
                    result += "        %s : %s;\n" % (CleanName(x[0]), CleanName(x[1]._containedType))
            else:
                result += "        %s : %s;\n" % (CleanName(x[0]), MapType_to_Ada(False, x, x[1], leafTypeDict, names))
        #    if isinstance(x[1], AsnSequenceOf) or \
        #       isinstance(x[1], AsnSetOf) or \
        #       isinstance(x[1], AsnString) or \
        #       (isinstance(x[1], AsnMetaMember) and \
        #           (isinstance(names[x[1]._containedType], AsnSequenceOf) or \
        #            isinstance(names[x[1]._containedType], AsnSetOf) or \
        #            isinstance(names[x[1]._containedType], AsnString))):
        #       result += "    " + CleanName(x[0]) + "_len : Integer;\n"
        result += "    end case;\n"
        result += "end record"
        g_choiceNodeTypename = ""
        return result
    else:  # pragma: no cover
        panic("MapType_to_Ada: Unsupported node type: %s" % realLeafType)  # pragma: no cover


def DeclareAdaTypesForC(nodeTypename, leafTypeDict, names, adaASN1TypesFile):
    if nodeTypename not in g_declared_Ada_Types:
        g_declared_Ada_Types[nodeTypename] = True
        # Unfortunately, Ada doesn't allow nameless structures.
        # We therefore have to scan all the nodeTypename subtree to find all the types
        # we will have to predefine...
        results = []
        ScanChildren(nodeTypename, names[nodeTypename], names, results, isRoot=True, createInnerNodesInNames=True)
        for prereqNodeTypename in results:
            if prereqNodeTypename != '':
                DeclareAdaTypesForC(prereqNodeTypename, leafTypeDict, names, adaASN1TypesFile)
        inform("Declaring Ada type: %s", nodeTypename)
        node = names[nodeTypename]
        # first, check if this type used to be an AsnMetaType (before VerifyAndFixAST made it a normal type)
        if nodeTypename in commonPy.asnParser.g_metatypes:
            if leafTypeDict[nodeTypename] == "INTEGER":
                adaASN1TypesFile.write('subtype %s is %s%s;\n\n' %
                                       (CleanName(nodeTypename),
                                        CleanName(commonPy.asnParser.g_metatypes[nodeTypename]),
                                        GetIntegerRangeAsAda(node)))
            else:
                adaASN1TypesFile.write('type %s is new %s;\n\n' %
                                       (CleanName(nodeTypename),
                                        CleanName(commonPy.asnParser.g_metatypes[nodeTypename])))
        else:
            if isinstance(node, AsnEnumerated):
                for x in node._members:
                    if x[1] is None:
                        panic("The Ada mapper needs integer values in the enumeration options (%s)" % node.Location())  # pragma: no cover
                adaASN1TypesFile.write('type %s is (' % CleanName(nodeTypename))
                adaASN1TypesFile.write(CleanName(node._members[0][0]))
                for m in node._members[1:]:
                    adaASN1TypesFile.write(', ')
                    adaASN1TypesFile.write(CleanName(m[0]))
                adaASN1TypesFile.write(');\n')
                adaASN1TypesFile.write('for %s use\n' % CleanName(nodeTypename))
                adaASN1TypesFile.write('    (%s => %s' % (CleanName(node._members[0][0]), node._members[0][1]))
                for m in node._members[1:]:
                    adaASN1TypesFile.write(', %s => %s' % (CleanName(m[0]), m[1]))
                adaASN1TypesFile.write(');\n')
                adaASN1TypesFile.write("for %s'Size use 32;\n\n" % CleanName(nodeTypename))
            else:
                if isinstance(node, AsnString) or isinstance(node, AsnSetOf) or isinstance(node, AsnSequenceOf):
                    adaASN1TypesFile.write('type %s_inner is\n' % CleanName(nodeTypename))
                elif isinstance(node, AsnChoice):
                    adaASN1TypesFile.write('type %s_selection is new Integer range 1..%d;\n' %
                                           (CleanName(nodeTypename),
                                            len(node._members)))
                    adaASN1TypesFile.write('type %s(choice_%s : %s_selection := 1) is\n' %
                                           (CleanName(nodeTypename),
                                            CleanName(nodeTypename),
                                            CleanName(nodeTypename)))
                    global g_choiceNodeTypename
                    g_choiceNodeTypename = nodeTypename
                else:
                    adaASN1TypesFile.write('type %s is\n' % CleanName(nodeTypename))
                if isinstance(node, AsnInt) or isinstance(node, AsnReal) or isinstance(node, AsnBool) or isinstance(node, AsnEnumerated):
                    adaASN1TypesFile.write('new ' + MapType_to_Ada(True, None, node, leafTypeDict, names))
                else:
                    adaASN1TypesFile.write(MapType_to_Ada(True, None, node, leafTypeDict, names))
                adaASN1TypesFile.write(';\n\n')
                if isinstance(node, AsnString) or isinstance(node, AsnSetOf) or isinstance(node, AsnSequenceOf):
                    adaASN1TypesFile.write('type %s is\n' % CleanName(nodeTypename))
                    adaASN1TypesFile.write('record\n')
                    adaASN1TypesFile.write('    length : Integer;\n')
                    adaASN1TypesFile.write('    data : %s_inner;\n' % CleanName(nodeTypename))
                    adaASN1TypesFile.write('end record;\n\n')
                elif isinstance(node, AsnChoice):
                    adaASN1TypesFile.write('for %s use\nrecord\n    choice_%s at 0 range 0..31;\nend record;\n\n' %
                                           (CleanName(nodeTypename),
                                            CleanName(nodeTypename)))


def CreateUnifiedAdaDeclarations(names, leafTypeDict, outputDir):
    adaASN1TypesFile = open(outputDir + "asn1_types.ads", "w")
    adaASN1TypesFile.write('package ASN1_Types is\n\n')

    adaASN1TypesFile.write('type Bool32bits is new Boolean;\n')
    adaASN1TypesFile.write("for Bool32bits'Size use 32;\n\n")

    adaASN1TypesFile.write('type Int64bits is new Long_Long_Integer;\n')  # range -9223372036854775808 .. 9223372036854775808;\n')
    adaASN1TypesFile.write("for Int64bits'Size use 64;\n\n")

    global g_declared_Ada_Types
    g_declared_Ada_Types = {}

#    dependsOn = {}
#    inform("Scanning dependency tree...")
#    for nodeTypename in names.iterkeys():
#       results = []
#       inform("Scanning %s..." % nodeTypename)
#       ScanChildren(nodeTypename, names[nodeTypename], names, results, isRoot=True, createInnerNodesInNames=True)
#       inform("%s depends on %s ..." % (nodeTypename, str(results)))
#       dependsOn[nodeTypename] = results[:]
#       assert(nodeTypename not in results)
#
#    inform("Scans completed")
#    inputQueue = names.iterkeys()
#    outputQueue = []
#    typesDoneSoFar = {}
#    while inputQueue:
#       for nodeTypename in inputQueue:
#           inform ("Checking %s" % nodeTypename)
#           inform ("OQ so far: %s" % str(outputQueue))
#           for dependency in dependsOn[nodeTypename]:
#               if not typesDoneSoFar.has_key(dependency):
#                   inform("No, %s" % dependency)
#                   break
#           else:
#               inform ("Output %s" % nodeTypename)
#               outputQueue.append(nodeTypename)
#               inputQueue.remove(nodeTypename)
#               typesDoneSoFar[nodeTypename]=1
#               break
#
#    for nodeTypename in outputQueue:
    for nodeTypename in names:
        DeclareAdaTypesForC(nodeTypename, leafTypeDict, names, adaASN1TypesFile)

    adaASN1TypesFile.write('end ASN1_Types;\n\n')
    adaASN1TypesFile.close()
