#!/usr/bin/env python

# (C) Semantix Information Technologies.
#
# Semantix Information Technologies is licensing the code of the
# Data Modelling Tools (DMT) in the following dual-license mode:
#
# Commercial Developer License:
#       The DMT Commercial Developer License is the suggested version
# to use for the development of proprietary and/or commercial software.
# This version is for developers/companies who do not want to comply
# with the terms of the GNU Lesser General Public License version 2.1.
#
# GNU LGPL v. 2.1:
#       This version of DMT is the one to use for the development of
# applications, when you are willing to comply with the terms of the
# GNU Lesser General Public License version 2.1.
#
# Note that in both cases, there are no charges (royalties) for the
# generated code.
#
'''
SMP2 Catalogues datatypes importer

This tool parses SMP2 Catalogues, extracts the data types described in them,
and maps them to the corresponding ASN.1 data type declarations. It also
includes logic to merge these types to pre-existing data type declarations
of an ASN.1 file - thus allowing merging of SMP2 designs into TASTE designs
(datatype-wise).
'''
import os
import re
import sys
import getopt
from lxml import etree

import commonPy.asnParser
from createInternalTypes import ScanChildren
from commonPy.asnAST import AsnBool, AsnInt, AsnReal, \
    AsnEnumerated, AsnOctetString, AsnSequenceOf, AsnSetOf, \
    AsnSequence, AsnSet, AsnChoice, AsnMetaMember, AsnMetaType

# Level of verbosity
g_verboseLevel = 0

# colors (used when calling 'info')
ESC = chr(27)
red = ESC+"[31m"
green = ESC+"[32m"
white = ESC+"[0m"
yellow = ESC+"[33m"
colors=[red, green, white, yellow]

# Lookup table for SMP2 types that map to AsnBasicNodes
simpleTypesTable = {
    'http://www.esa.int/2005/10/Smp#Bool': (AsnBool, None, None),
    'http://www.esa.int/2005/10/Smp#Int8': (AsnInt, -128, 127),
    'http://www.esa.int/2005/10/Smp#Int16': (AsnInt, -32768, 32767),
    'http://www.esa.int/2005/10/Smp#Int32': (AsnInt, -2147483648, 2147483647),
    'http://www.esa.int/2005/10/Smp#Int64': (AsnInt, None, None),
    'http://www.esa.int/2005/10/Smp#UInt8': (AsnInt, 0, 255),
    'http://www.esa.int/2005/10/Smp#UInt16': (AsnInt, 0, 65535),
    'http://www.esa.int/2005/10/Smp#UInt32': (AsnInt, 0, 4294967295),
    'http://www.esa.int/2005/10/Smp#UInt64': (AsnInt, None, None),
    'http://www.esa.int/2005/10/Smp#Float32': (AsnReal, None, None),
    'http://www.esa.int/2005/10/Smp#Float64': (AsnReal, None, None)
}


def info(level, *args):
    '''Checks the 'level' argument against g_verboseLevel and then prints
    the rest of the args, one by one, separated by a space. It also
    has logic to deal with usage of one of the colors as arguments
    (in which case it avoids printing spurious spaces).
    '''
    if not args:
        panic("You called info without args")  # pragma: no cover
    if level<=g_verboseLevel:
        for i in xrange(len(args)):
            if i !=0 and args[i-1] not in colors:
                sys.stdout.write(' ')
            sys.stdout.write(args[i])
        for i in xrange(len(args)-1, -1, -1):
            if args[i] in colors:
                continue
            if not args[i].endswith('\n'):
                sys.stdout.write('\n')
                return


def panic(x, coloredBanner=""):
    '''Notifies the user that something fatal happened and aborts. '''
    info(0, yellow + coloredBanner + white + '\n' + x)
    sys.exit(1)


def usage(coloredMsg=""):
    '''Prints help message and aborts. '''
    usageMsg = 'Usage: smp2asn <options> <smp2Catalogues...>\n\n' \
        'Where options must include:\n' \
        '  -o, --outAsn1=newAsnGrammar.asn  the output ASN.1 grammar, containing\n'\
        '                                   all the SMP2 types (if -a was used,\n'\
        '                                   the existing ASN.1 types, too)\n'\
        'Options may also include:\n'\
        '  -a, --asn1=asnGrammar.asn        an input ASN.1 grammar to merge with\n'\
        '  -p, --prune                      prune unnamed (inner) SMP2-translation types\n'\
        '  -v, --verbose                    Be more verbose (debugging)\n' \
        '  -h, --help                       Show this help message\n'
    panic(usageMsg, coloredMsg)


def Clean(fieldName):
    '''When mapping field names and type names from SMP2 to ASN.1,
    we need to change '_' to '-'. '''
    return re.sub(r'[^a-zA-Z0-9-]', '-', fieldName)


class DashUnderscoreAgnosticDict(dict):
    '''A dictionary that automatically replaces '_' to '-' in its keys. '''
    def __setitem__(self, key, value):
        super(DashUnderscoreAgnosticDict, self).__setitem__(key.replace('_', '-'), value)

    def __getitem__(self, key):
        return super(DashUnderscoreAgnosticDict, self).__getitem__(key.replace('_', '-'))

    def __contains__(self, key):
        return super(DashUnderscoreAgnosticDict, self).__contains__(key.replace('_', '-'))


class Attributes:
    '''Helper class, to ease access to XML attributes.
    It allows us to write code like this...

            a = Attributes(lxmlEtreeNode)
            whatever = a.href
            print a.title

        ...instead of this:

            whatever = lxmlEtreeNode.get('href', None)
            print a.get('title', None)
    '''
    def __init__(self, t):
        '''Argument t is an lxml Etree node.'''
        self._attrs = {}
        for k, v in t.items():
            endBraceIdx = k.find('}')
            if endBraceIdx != -1:
                k = k[endBraceIdx+1:]
            self._attrs[k] = v

    def __getattr__(self, x):
        return self._attrs.get(x, None)


def MapSMP2Type(attrs, enumOptions, itemTypes, fields):
    '''
    Core mapping function. Works on the XML attributes of the lxml Etree node,
    and returns a node from commonPy.asnAST.
    '''
    location = 'from %s, in line %s' % (attrs.base, attrs.sourceline)
    info(2, "Mapping SMP2 type", location)

    def getMaybe(cast, x):
        try:
            return cast(x)
        except:  # pragma: no cover
            return None  # pragma: no cover
    dataDict = {"asnFilename": attrs.base, "lineno": attrs.sourceline}
    if attrs.type == 'Types:Integer':
        low = getMaybe(int, attrs.Minimum)
        high = getMaybe(int, attrs.Maximum)
        if low==0 and high==1:
            # Pseudo-boolean from TASTE mapping, as per SpaceBel instructions
            return AsnBool(**dataDict)
        else:
            # Normal integer
            span = [low, high] if low is not None and high is not None else []
            dataDict["range"] = span
            return AsnInt(**dataDict)
    elif attrs.type == 'Types:Float':
        low = getMaybe(float, attrs.Minimum)
        high = getMaybe(float, attrs.Maximum)
        span = [low, high] if low is not None and high is not None else []
        dataDict["range"] = span
        return AsnReal(**dataDict)
    elif attrs.type == 'Types:Enumeration':
        dataDict["members"] = enumOptions
        return AsnEnumerated(**dataDict)
    elif attrs.type == 'Types:String':
        high = getMaybe(int, attrs.Length)
        span = [high, high] if high is not None else []
        dataDict["range"] = span
        return AsnOctetString(**dataDict)
    elif attrs.type == 'Types:Array':
        if itemTypes == []:
            panic("Missing mandatory ItemType element", location)  # pragma: no cover
        itemTypeAttrs = Attributes(itemTypes[0])
        arrSize = getMaybe(int, attrs.Size)
        if not arrSize:
            panic("Missing array 'Size' attribute", location)  # pragma: no cover
        dataDict["range"] = [arrSize, arrSize]
        if itemTypeAttrs.href in [
                'http://www.esa.int/2005/10/Smp#Char8',
                'http://www.esa.int/2005/10/Smp#Int8',
                'http://www.esa.int/2005/10/Smp#UInt8']:
            return AsnOctetString(**dataDict)
        else:
            containedHref = itemTypeAttrs.href
            if not containedHref:
                panic("Missing reference to 'href' (file:%s, line:%d)" %
                      itemTypeAttrs.base, itemTypeAttrs.sourceline)  # pragma: no cover
            idxHash = containedHref.find('#')
            if -1 != idxHash:
                containedHref = containedHref[idxHash+1:]
            if itemTypeAttrs.href in simpleTypesTable.keys():
                # Create the AsnBasicNode this child maps to.
                cast, low, high = simpleTypesTable[itemTypeAttrs.href]
                span=[low, high] if low and high else []
                childDict = {
                    'asnFilename': itemTypes[0].base,
                    'lineno': itemTypes[0].sourceline
                }
                if span != []:
                    childDict['range'] = span
                childNode = cast(**childDict)
                dataDict['containedType'] = childNode
            else:
                # Store the 'Id' attribute - we will resolve this
                # in the FixupOutOfOrderIdReferences function.
                dataDict['containedType'] = containedHref
            return AsnSequenceOf(**dataDict)
    elif attrs.type == 'Types:Structure':
        members = []
        for field in fields:
            try:
                fieldName = field.get('Name')
                if fieldName != 'choiceIdx':
                    fieldName = Clean(fieldName)
                    fieldName = fieldName[0].lower() + fieldName[1:]
                    try:
                        refTypeAttrs = Attributes(field.xpath("Type")[0])
                    except:  # pragma: no cover
                        location = 'from %s, in line %s' % \
                            (field.base, field.sourceline)  # pragma: no cover
                        panic("Missing Type child element", location)  # pragma: no cover
                    refTypeHref = refTypeAttrs.href
                    idxHash = refTypeHref.find('#')
                    if -1 != idxHash:
                        refTypeHref = refTypeHref[idxHash+1:]
                    if refTypeAttrs.href in simpleTypesTable.keys():
                        cast, low, high = simpleTypesTable[refTypeAttrs.href]
                        containedDict = {
                            'asnFilename': field.base,
                            'lineno': field.sourceline
                        }
                        span = [low, high] if low and high else []
                        if span != []:
                            containedDict['range']=[low, high]
                        basicNode = cast(**containedDict)
                        members.append((fieldName, basicNode))
                    else:
                        members.append((fieldName, AsnMetaMember(
                            asnFilename=field.base,
                            lineno=field.sourceline,
                            containedType=refTypeHref)))
                else:
                    members.append((fieldName, 'dummy'))
            except Exception, e:  # pragma: no cover
                panic(str(e) + '\nMake sure that:\n'
                      '1. The "Name" attribute exists\n'
                      '2. The "Type" child element, with attribute '
                      '"xlink:title" also exists.',
                      'In %s, line %d:' % (field.base, field.sourceline))  # pragma: no cover
        if 0 == len(members):
            panic("Empty SEQUENCE is not supported", location)  # pragma: no cover
        if members[0][0] == 'choiceIdx':
            dataDict['members'] = members[1:]
            return AsnChoice(**dataDict)
        else:
            dataDict['members'] = members
            return AsnSequence(**dataDict)
    panic("Failed to map... (%s)" % attrs.type, location)  # pragma: no cover


def FixupOutOfOrderIdReferences(nodeTypename, asnTypesDict, idToTypeDict):
    '''Based on the uniqueness of the 'Id' elements used in
    'xlink:href' remote references, we resolve the lookups of
    remote types that we stored in AsnMetaMembers during MapSMP2Type().'''
    node = asnTypesDict[nodeTypename]
    if isinstance(node, AsnChoice) or isinstance(node, AsnSequence) or isinstance(node, AsnSet):
        for idx, child in enumerate(node._members):
            if isinstance(child[1], AsnMetaMember):
                containedType = child[1]._containedType
                if containedType in idToTypeDict:
                    containedType = idToTypeDict[containedType]
                if containedType in asnTypesDict:
                    node._members[idx] = (child[0], asnTypesDict[containedType])
                else:
                    panic("Could not resolve Field '%s' in type '%s' (contained: %s)..." %
                          (child[0], nodeTypename, containedType), node.Location())  # pragma: no cover
    elif isinstance(node, AsnSequenceOf) or isinstance(node, AsnSetOf):
        if isinstance(node._containedType, str):
            containedType = node._containedType
            if containedType in idToTypeDict:
                containedType = idToTypeDict[containedType]
            if containedType in asnTypesDict:
                node._containedType = asnTypesDict[containedType]
            else:
                panic("Could not resolve Field '%s' in type '%s'... (contained: %s)" %
                      (child[0], nodeTypename, containedType), node.Location())  # pragma: no cover


def ConvertCatalogueToASN_AST(inputSmp2Files):
    '''Converts a list of input SMP2 Catalogues into an ASN.1 AST,
    which it returns to the caller.'''
    asnTypesDict = DashUnderscoreAgnosticDict()
    idToTypeDict = {}
    allSMP2Types = {}
    # Do a first pass, verifying the primary assumption:
    # That 'Id' elements of types are unique across our set of SMP2 files.
    for inputSmp2File in inputSmp2Files:
        a=etree.parse(open(inputSmp2File))
        root=a.getroot()
        if len(root)<1 or not root.tag.endswith('Catalogue'):
            panic('', "You must use an XML file that contains an SMP2 Catalogue")  # pragma: no cover
        for t in root.xpath("Namespace/Type"):
            a = Attributes(t)
            if a.Id in allSMP2Types:
                catalogue = allSMP2Types[a.Id]  # pragma: no cover
                if catalogue != inputSmp2File:  # pragma: no cover
                    panic("The same Id exists in two files: %s exists in both: %s" %
                          (a.Id, str([catalogue, inputSmp2File])))  # pragma: no cover
            else:
                allSMP2Types[a.Id] = inputSmp2File
    for inputSmp2File in inputSmp2Files:
        a=etree.parse(open(inputSmp2File))
        root=a.getroot()
        if len(root)<1 or not root.tag.endswith('Catalogue'):
            panic('', "You must use an XML file that contains an SMP2 Catalogue")  # pragma: no cover
        for t in root.xpath("Namespace/Type"):
            namespace = t.getparent()
            nsName = namespace.get('Name')
            if not nsName:
                panic("Missing attribute Name from Namespace (file:%s, line:%d)" %
                      namespace.base, namespace.sourceline)  # pragma: no cover
            cataloguePrefix = Clean(nsName).capitalize() + "_"

            a = Attributes(t)
            a.base = t.base
            a.sourceline = t.sourceline

            if not a.type or not a.Name:
                panic("Both 'xsi:type' and 'Name' are mandatory attributes (file:%s, line:%d)" %
                      a.base, a.sourceline)  # pragma: no cover
            if a.type.startswith('Catalogue:'):
                # We only wants Types, nothing more
                continue

            nodeTypename = a.Name
            nodeTypename = nodeTypename[0].upper() + nodeTypename[1:]
            nodeTypename = nodeTypename.replace('_', '-')

            # Gather children node's info:

            # 1. Enumeration data
            enumOptions = []
            if a.type == 'Types:Enumeration':
                for node in t.xpath("Literal"):
                    enumOptions.append(map(node.get, ['Name', 'Value']))

            # 2. ItemType data (used in arrays)
            itemTypes = t.xpath("ItemType")

            # 3. Field data (used in structures)
            fields = t.xpath("Field")

            info(2, "Creating type:", cataloguePrefix+nodeTypename)
            asnTypesDict[cataloguePrefix + nodeTypename] = \
                MapSMP2Type(a, enumOptions, itemTypes, fields)
            # Store mapping from Id to typename in idToTypeDict
            # (used below, in FixupOutOfOrderIdReferences)
            idToTypeDict[a.Id] = cataloguePrefix + nodeTypename

    for nodeTypename in asnTypesDict.keys():
        FixupOutOfOrderIdReferences(nodeTypename, asnTypesDict, idToTypeDict)
    return asnTypesDict


def MergeASN1_AST(smp2AsnAST):
    '''Merges the ASN.1 AST generated from SMP2 files (smp2AsnAST param)
    into the ASN.1 AST stored in commonPy.asnParser.g_names. Uses smart
    merging, i.e. diff-like semantics.'''
    typesToAddVerbatim = []
    identicals = {}
    d = commonPy.asnParser.g_names
    for k, v in smp2AsnAST.iteritems():
        if k in d:
            # Type name exists in both trees - is it the same?
            if not v.IdenticalPerSMP2(d[k], smp2AsnAST, d):  # pragma: no cover
                panic(green + k + white + " exists, but is different:\n" +
                      "it is...\n" + d[k].AsASN1(d) + "\n" +
                      "but in SMP2 it is...\n" + v.AsASN1(smp2AsnAST))  # pragma: no cover
            else:  # pragma: no cover
                info(1, green, k, white, "exists and is semantically equivalent.")  # pragma: no cover
        else:
            # Find an identical type if possible
            for k2, v2 in d.iteritems():
                if v2._isArtificial:
                    # Avoid mapping to artificially generated inner types
                    # (see last part of VerifyAndFixAST in commonPy.asnParser)
                    continue
                if v2.IdenticalPerSMP2(v, d, smp2AsnAST):
                    info(1, green, k, white, "is identical to", red, k2, white)
                    identicals[k] = k2
                    break
            else:
                info(1, green, k, white, "must be copied (no equivalent type found)...")
                typesToAddVerbatim.append(k)
    # Merge missing types in commonPy.asnParser.g_names
    for nodeTypename in typesToAddVerbatim:
        results = []
        node = smp2AsnAST[nodeTypename]
        # Take care to add dependencies first
        ScanChildren(nodeTypename, node, smp2AsnAST, results, isRoot=True, createInnerNodesInNames=False)
        info(1, "Will copy", nodeTypename, "("+str(node.__class__)+")", ("and "+str(results) if results else ''))
        results.append(nodeTypename)
        for r in results:
            node = smp2AsnAST[r]
            d[r] = node
            if isinstance(node, AsnMetaType):
                commonPy.asnParser.g_metatypes[r] = r._containedType  # pragma: no cover
                d[r] = smp2AsnAST[r._containedType]  # pragma: no cover
            commonPy.asnParser.g_typesOfFile.setdefault(node._asnFilename, []).append(r)
    return identicals


def SaveASN_AST(bPruneUnnamedInnerTASTEtypes, outputAsn1Grammar, identicals):
    d = DashUnderscoreAgnosticDict()
    for k, v in commonPy.asnParser.g_names.iteritems():
        d[k] = v
    with open(outputAsn1Grammar, 'w') as f:
        f.write('DATAVIEW DEFINITIONS AUTOMATIC TAGS ::= BEGIN\n\n')
        for k, v in d.iteritems():
            if v._isArtificial:
                # Don't emit artificially generated inner types
                # (see last part of VerifyAndFixAST in commonPy.asnParser)
                continue
            if bPruneUnnamedInnerTASTEtypes and 'TaStE' in k:
                # Don't emit artificially generated SMP2 types
                # (see smp2_A_mapper)
                continue
            f.write('-- From ' + v._asnFilename + ' line ' + str(v._lineno) + '\n')
            f.write(k + ' ::= ')
            f.write(v.AsASN1(d) + "\n\n")
        for k, v in identicals.iteritems():
            f.write(k + ' ::= ' + v + '\n\n')
        f.write('END\n')


def main(args):
    try:
        optlist, args = getopt.gnu_getopt(
            args,
            "hvpi:a:o:", ['help', 'verbose', 'prune', 'smp2=', 'asn1=', 'outAsn1='])
    except:
        usage("Invalid parameters passed...")
    inputSmp2Files = args
    inputAsn1Grammar = ""
    outputAsn1Grammar = ""
    bPrune = False
    for opt, arg in optlist:
        if opt in ("-h", "--help"):
            usage()
        elif opt in ("-a", "--asn1"):
            inputAsn1Grammar = arg
        elif opt in ("-o", "--outAsn1"):
            outputAsn1Grammar = arg
        elif opt in ("-p", "--prune"):
            bPrune = True
        elif opt in ("-v", "--verbose"):
            global g_verboseLevel
            g_verboseLevel += 1

    if not inputSmp2Files:
        usage("SMP2 catalogues were not provided")
    if not outputAsn1Grammar:
        usage("Mandatory option (-o newAsnGrammar.asn) missing")

    def CheckFileExists(x):
        if not os.path.isfile(os.path.realpath(x)):
            usage("Input (%s) is not a file" % x)
    map(CheckFileExists, inputSmp2Files)
    if inputAsn1Grammar:
        CheckFileExists(inputAsn1Grammar)

    smp2AsnAST = ConvertCatalogueToASN_AST(inputSmp2Files)
    if inputAsn1Grammar:
        commonPy.asnParser.ParseAsnFileList([inputAsn1Grammar])
    identicals = MergeASN1_AST(smp2AsnAST)
    SaveASN_AST(bPrune, outputAsn1Grammar, identicals)
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
