-------------------------------------------------------------------------------------------------
-- Files       : Wrapper.vhd          
-- Version     : 0.1                        
-- Date        : 07/06/2010                 
-- Author      : Yann LECLERC - M3Systems   
-------------------------------------------------------------------------------------------------
-- Description :
-- Wrapper between APB bus and custom IP, one entity and architecture declaration by custom IP.
-- This wrapper is clocked, with asynchronous reset (active low).
-- The configuration of this IP on the APB Bus is transmited throught the pconfig constant,
-- and is defined as follow :   
-- Vendor ID : 0xff
-- Device ID : 0x001
-- Version ID : 0x001
-- Interrupt index : 0x0
-- APB slave index : defined by pindex generic
-- APB address : defined by paddr generic
-- Size of the address space : defined by pmask (always 0xfff to defined 256 bytes memory space)
--
-- This wrapper is composed of a set of 64 32 bits registers (the 256 bytes memory space), and
-- the custom IP. 
-- The registers are divided into 3 parts : 
-- 1)The control and command register at offset 0. The bit 0 of this register drives the 
-- start_compute signal, and can be read or write. The bit 1 is read only, and is driven by the 
-- compute signal. The bit 2 is read only, and is driven by the finish_compute signal.
-- 2)The read and write registers area. Each register in this area is linked to custom IP's input
-- variables, and can be read or written. The address range of this area is defined by the 
-- size_rw_area constant as follow : from offset 1 to offset size_rw_area.
-- 3)The read only registers area. Each register in this area is linked to custom IP's output
-- variables, and can only be read. The address range of this area rely on the size of the 
-- custom IP output variables.
--
-- The start_compute is linked to the custom IP, and used to indicate that the computation can
-- start. The finish_compute signal is linked to the custom IP, and must be used to indicate that
-- the computation is finish. The compute signal is a combination of start_compute and 
-- finish_compute signal.
--
-- In order to implement a custom IP a wrapper entity and architecture declaration must be added, 
-- with the following elements that must be changed : 
-- The custom IP component declaration and instanciation. The size of the read and write 
-- registers area. The link between custom IP input variables and read and write registers. The 
-- link between custom IP output variables and read only registers.
-------------------------------------------------------------------------------------------------

--apbwrapper for customip0 declaration--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library grlib;
use grlib.amba.all;

entity apbwrapper0 is
	generic(
		--Configuration information--
		pindex : integer;
		paddr : integer;
		pmask : integer
		);
	port(
		--Bus signals--
		apbi : in  apb_slv_in_type;
		apbo : out apb_slv_out_type;
		
		rst : in std_logic;
		
		clk : in std_logic
		
		);
end apbwrapper0;

architecture archiapbwrapper0 of apbwrapper0 is

	--Configuration declaration--
	constant pconfig : apb_config_type := (
		0 => ahb_device_reg ( 16#ff#, 1, 0, 1, 0),
		1 => apb_iobar(paddr, pmask));
	
	--Type--	
	type registers_type is array(63 downto 0) of std_logic_vector(31 downto 0);
	
	--Signals--
	signal start_compute : std_logic;
	signal finish_compute : std_logic;
	
	--Size of the read and write registers area--
	constant size_rw_area : integer := 2;
	
	--CustomIP Inputs--
	signal custom_var_in_1 : std_logic_vector(31 downto 0);
	signal custom_var_in_2 : std_logic_vector(31 downto 0);
	
	--CustomIP Outputs--
	signal custom_var_out_1 : std_logic_vector(31 downto 0);
	signal custom_var_out_2 : std_logic_vector(31 downto 0);
	
	--CustomIP component--
	component customip0 is
	port(
		--Input signals--
		custom_var_in_1 : in std_logic_vector(31 downto 0);
		custom_var_in_2 : in std_logic_vector(31 downto 0);
		
		--Output signals--
		custom_var_out_1 : out std_logic_vector(31 downto 0);
		custom_var_out_2 : out std_logic_vector(31 downto 0);
		
		--Control signals--
		start_compute : in std_logic;
		finish_compute : out std_logic;
		
		rst : in std_logic;
		
		clk_compute : in std_logic
		);
	end component;
	
begin

	--APB Slave Configuration--
	apbo.pconfig <= pconfig;

	--CustomIP instanciation--
	cip : customip0 port map(custom_var_in_1,custom_var_in_2,
							custom_var_out_1,custom_var_out_2,
							start_compute,finish_compute,
							rst,clk);
	
	--Read and write process--
	read_write : 
	process(clk,rst)
		variable sel : std_logic;
		variable en : std_logic;
		variable wr : std_logic;
		variable add : std_logic_vector(7 downto 0);
		variable ind : integer;
		variable data_in : std_logic_vector(31 downto 0);
		variable regs : registers_type;
	begin
		if rst='0' then -- Asynchronous Reset
		
			sel := '0';
			en := '0';
			wr := '0';
			add := x"00";
			ind := 0;
			data_in := x"00000000";
			regs := (others => (others => '0'));
			start_compute <= '0';
			
		elsif clk'event and clk='1' then
		
			-- AMBA bus signals extraction --
			sel := apbi.psel(pindex); -- APB slave selected
			en := apbi.penable; -- APB Slave action enabled
			wr := apbi.pwrite; -- Write action
			add := apbi.paddr(7 downto 0); -- 256 bytes address
			data_in :=  apbi.pwdata; -- Receipt data to write
			
			ind := conv_integer(add(7 downto 2)); -- Registers index, read and write are possible only by step of 32 bits
			
			regs(0)(2) := finish_compute; -- Finish_compute is linked to bit 2 of Control/Command register
			regs(0)(1) := regs(0)(0) and (not regs(0)(2)); -- Computation is running if start_compute is true and finish_compute is true  
			
			-- Link Custom IP output variables with registers -- 
			regs(3) := custom_var_out_1;
			regs(4) := custom_var_out_2;
			
			apbo.prdata <= regs(ind); -- Read
			
			if (sel and (wr) and en) ='1' then -- Write
				if ind = 0 then -- Control/Command register
					regs(0)(0) := data_in(0);
				elsif ind < size_rw_area+1 then -- Read and write registers area from 1 to size_rw_area
					regs(ind) := data_in;
				end if;
			end if;
			
			start_compute <= regs(0)(0); -- Drive start_compute
			
			-- Link Custom IP input variables with registers --
			custom_var_in_1 <= regs(1);
			custom_var_in_2 <= regs(2);
			
		end if;
	end process;

	
end archiapbwrapper0;

--apbwrapper for customip1 declaration--
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library grlib;
use grlib.amba.all;

entity apbwrapper1 is
	generic(
		--Configuration information--
		pindex : integer;
		paddr : integer;
		pmask : integer
		);
	port(
		--Bus signals--
		apbi : in  apb_slv_in_type;
		apbo : out apb_slv_out_type;
		
		rst : in std_logic;
		
		clk : in std_logic
		
		);
end apbwrapper1;

architecture archiapbwrapper1 of apbwrapper1 is

	--Configuration declaration--
	constant pconfig : apb_config_type := (
		0 => ahb_device_reg ( 16#ff#, 1, 0, 1, 0),
		1 => apb_iobar(paddr, pmask));
	
	--Type--	
	type registers_type is array(63 downto 0) of std_logic_vector(31 downto 0);
	
	--Signals--
	signal start_compute : std_logic;
	signal finish_compute : std_logic;
	
	--Size of the read and write registers area--
	constant size_rw_area : integer := 2;
	
	--CustomIP Inputs--
	signal custom_var_in_1 : std_logic_vector(31 downto 0);
	signal custom_var_in_2 : std_logic_vector(31 downto 0);
	
	--CustomIP Outputs--
	signal custom_var_out_1 : std_logic_vector(31 downto 0);
	signal custom_var_out_2 : std_logic_vector(31 downto 0);
	
	--CustomIP component--
	component customip1 is
	port(
		--Input signals--
		custom_var_in_1 : in std_logic_vector(31 downto 0);
		custom_var_in_2 : in std_logic_vector(31 downto 0);
		
		--Output signals--
		custom_var_out_1 : out std_logic_vector(31 downto 0);
		custom_var_out_2 : out std_logic_vector(31 downto 0);
		
		--Control signals--
		start_compute : in std_logic;
		finish_compute : out std_logic;
		
		rst : in std_logic;
		
		clk_compute : in std_logic
		);
	end component;
	
begin

	--APB Slave Configuration--
	apbo.pconfig <= pconfig;

	--CustomIP instanciation--
	cip : customip1 port map(custom_var_in_1,custom_var_in_2,
							custom_var_out_1,custom_var_out_2,
							start_compute,finish_compute,
							rst,clk);
	
	--Read and write process--
	read_write : 
	process(clk,rst)
		variable sel : std_logic;
		variable en : std_logic;
		variable wr : std_logic;
		variable add : std_logic_vector(7 downto 0);
		variable ind : integer;
		variable data_in : std_logic_vector(31 downto 0);
		variable regs : registers_type;
	begin
		if rst='0' then -- Asynchronous Reset
		
			sel := '0';
			en := '0';
			wr := '0';
			add := x"00";
			ind := 0;
			data_in := x"00000000";
			regs := (others => (others => '0'));
			start_compute <= '0';
			
		elsif clk'event and clk='1' then
		
			-- AMBA bus signals extraction --
			sel := apbi.psel(pindex); -- APB slave selected
			en := apbi.penable; -- APB Slave action enabled
			wr := apbi.pwrite; -- Write action
			add := apbi.paddr(7 downto 0); -- 256 bytes address
			data_in :=  apbi.pwdata; -- Receipt data to write
			
			ind := conv_integer(add(7 downto 2)); -- Registers index, read and write are possible only by step of 32 bits
			
			regs(0)(2) := finish_compute; -- Finish_compute is linked to bit 2 of Control/Command register
			regs(0)(1) := regs(0)(0) and (not regs(0)(2)); -- Computation is running if start_compute is true and finish_compute is true  
			
			-- Link Custom IP output variables with registers -- 
			regs(3) := custom_var_out_1;
			regs(4) := custom_var_out_2;
			
			apbo.prdata <= regs(ind); -- Read
			
			if (sel and (wr) and en) ='1' then -- Write
				if ind = 0 then -- Control/Command register
					regs(0)(0) := data_in(0);
				elsif ind < size_rw_area+1 then -- Read and write registers area from 1 to size_rw_area
					regs(ind) := data_in;
				end if;
			end if;
			
			start_compute <= regs(0)(0); -- Drive start_compute
			
			-- Link Custom IP input variables with registers --
			custom_var_in_1 <= regs(1);
			custom_var_in_2 <= regs(2);
			
		end if;
	end process;

	
end archiapbwrapper1;

